/*========================================================================
 * FILE:        $RCSfile: Page.java,v $
 *
 * AUTHOR:      Nathan Waite
 * DATE:        $Date: 1998/08/06 16:47:17 $
 * REVISION:    $Revision: 1.1 $
 *
 * DESCRIPTION: Page class to support basic page storage and I/O needs.
 *
 * $Log: Page.java,v $
 * Revision 1.1  1998/08/06  16:47:17  liddle
 * Initial revision
 *
 *------------------------------------------------------------------------
 * Pattern Editor
 * Data Extraction Research Group
 * School of Accountancy and Information Systems
 * Marriott School of Management
 * 540 TNRB
 * Brigham Young University
 * Provo, UT 84602-3068
 * liddle@byu.edu
 *
 * Copyright (C) 1998, Board of Trustees of Brigham Young University
 *
 * Pattern Editor software, both binary and source (hereafter, Software)
 * is copyrighted by The Board of Trustees of Brigham Young University
 * (BYU), and ownership remains with BYU.
 *
 * BYU grants you (hereafter, Licensee) a license to use the Software
 * for academic, research and internal business purposes only, without a
 * fee.  Licensee may distribute the binary and source code (if released)
 * to third parties provided that the copyright notice and this statement
 * appears on all copies and that no charge is associated with such
 * copies.
 *
 * Licensee may not make derivative works without first arranging an
 * additional license agreement with BYU.
 *
 * Any Licensee wishing to make commercial use of the Software should
 * contact BYU to negotiate an appropriate license for such
 * commercial use.  Commercial use includes (1) integration of all or
 * part of the source code into a product for sale or license by or on
 * behalf of Licensee to third parties, or (2) distribution of the binary
 * code or source code to third parties that need it to utilize a
 * commercial product sold or licensed by or on behalf of Licensee.
 *
 * BYU MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE
 * FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED
 * WARRANTY.  BYU SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE
 * USERS OF THIS SOFTWARE.
 *
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the United States
 * including, but not limited to, export control laws, and the terms
 * of this license.  BYU shall have the right to terminate this license
 * immediately by written notice upon Licensee's breach of, or
 * non-compliance with, any of its terms.  Licensee may be held legally
 * responsible for any copyright infringement that is caused or
 * encouraged by Licensee's failure to abide by the terms of this license.
 *
 * Comments and questions are welcome and can be sent to liddle@byu.edu.
 */
/*------------------------------------------------------------------------
 *                      IMPORTS
 */
import java.io.*;

/*========================================================================
 *                      Page CLASS
 */
public class Page {

/*------------------------------------------------------------------------
 *                      PUBLIC CONSTANTS
 */
public static final int        	MAX_ROWS = 100,
				MAX_COLUMNS = 200;

/*------------------------------------------------------------------------
 *                      PRIVATE CONSTANTS
 */
private static final char       CARRIAGE_RETURN = '\r',
				END_OF_PAGE_CHAR = '\f',
				END_OF_ROW_CHAR = '\n';
private static final byte       END_OF_PAGE_BYTE = (byte)'\f',
				END_OF_ROW_BYTE = (byte)'\n';
private static final String     END_OF_PAGE_STR = "\f",
				END_OF_ROW_STR = "\n";

/*------------------------------------------------------------------------
 *                      ATTRIBUTES
 */
private char[][]        Content         = new char[MAX_ROWS][MAX_COLUMNS];
private int[]           Row_width       = new int [MAX_ROWS];
private int             Max_row_width   = 0,
			Number_of_rows  = 0;
private PrintWriter	Debug_out	= null;
// NEEDSWORK: add more messages for tracing in debug mode.

/*------------------------------------------------------------------------
 *                      PUBLIC CONSTRUCTORS
 */

/**
 * Constructs an empty page.
 */
public
Page()
{
  Clear();

} /* Page default constructor */

/**
 * Constructs self as a copy of a given Page.
 *
 * @param to_copy The Page from which self will be copied.
 */
public
Page(Page to_copy)
{
  Clear();
  Copy(to_copy);

} /* Page copy constructor */

/**
 * Constructs self from a given DataInputStream.  If there is no more
 * data in the DataInputStream, the self will be an empty Page.
 *
 * @param in The DataInputStream from which self will be initialized.
 */
public
Page(DataInputStream din)
{
  Clear();
  Read(din);

} /* Page DataInputStream constructor */

/*========================================================================
 *                      PUBLIC METHODS
 */
/*------------------------------------------------------------------------
 *                      PUBLIC MUTATORS
 */
/**
 * Clear the contents of self to spaces.
 */
public void
Clear()
{
  for (int i = 0; i < MAX_ROWS; i++) {
    for (int j = 0; j < MAX_COLUMNS; j++) {
      Content[i][j] = ' ';
    }
    Row_width[i] = 0;
  }
  Max_row_width = 0;
  Number_of_rows = 0;

} /* Clear */

/**
 * Set the character at a given cell on this Page.
 *
 * @param row 0-based row index of the cell to modify.
 * @param col 0-based column index of the cell to modify.
 * @param ch  Character to assign in the given cell.
 */
public void
SetChar(int row, int col, char ch)
{
  if (Character.isSpaceChar(ch)) {
    // Deleting from the page.
    Content[row][col] = ' ';
    if (col + 1 >= Row_width[row] && (row < 0))
      while (Content[row][Row_width[row] - 1] == ' ')
        --Row_width[row];
    if (Row_width[row] <= 0 && row + 1 >= Number_of_rows)
      while (Number_of_rows > 0 && Row_width[Number_of_rows - 1] <= 0)
        --Number_of_rows;
  } else {
    // Adding to the page.
    Content[row][col] = ch;
    if (row >= Number_of_rows) {
      Number_of_rows = row + 1;
    }
    if (col >= Row_width[row]) {
      Row_width[row] = col + 1;
      if (col + 1 > Max_row_width)
        Max_row_width = col + 1;
    }
  }

} /* SetChar */

/*
 * NEEDSWORK: Please check nested if to see if it handles multiple
 * lines correctly!
 *
 * SWL:  I'd rather go with a single line restriction.  Terminate the copy
 * when we hit the physical end of line, or when the input string
 * has a \n or \f or end of string.
 */
/**
 * Set the content at a given row and column for the length on this Page.
 * Length may not overlap onto multiple rows.
 *
 * @param row Row index of the string to modify.
 * @param col Column index of the string to modify.
 * @param len Length of the string to modify.
 * @param to_copy String to copy at the given row and column and length.
 */
public void
SetRange(int row, int col, int len, String to_copy)
{
  char[]        ch_assign;

  ch_assign = to_copy.toCharArray();
  for (int i = 0; i < len; i++) {
    Content[row][i + col] = MakePrintable(ch_assign[i]);
    if ((i + col) >= MAX_COLUMNS) {
      row++;
      len -= i;
      i = 0;
      col = 0;
    }
  }

} /* SetRange */

/**
 * Set the content at a given row and column on this Page.  Copies
 * the given String until either end of row or end of string.
 *
 * @param row 0-based row index of the area to modify.
 * @param col 0-based column index of the area to modify.
 * @param to_copy String to copy at the given row and column.
 */
public void
SetRange(int row, int col, String to_copy)
{
  char[] ch_assign;

  ch_assign = to_copy.toCharArray();
  for (int i = 0; i < MAX_COLUMNS - col; i++) {
    /*
     * NEEDSWORK: If the input character is \n or \f, terminate.
     *            If the input is not printable ASCII (32 to 126),
     *            don't copy it, but use some other character,
     *            like '.' instead.  We ought to make a facilitator
     *            called MakePrintable that receives a character,
     *            and checks it, then returns the original char if
     *            it is printable, or '.' otherwise.  It could print
     *            a message or throw an exception if it ever receives
     *            a non-printable character.  MakePrintable would
     *            need to be called here and elsewhere when we assign
     *            new characters to Content.
     *
     *            I suppose the other thing we ought to check is
     *            whether the given String is empty, in which case
     *            we're deleting from the page, not adding to it.
     *            Let's work on this after we have everything else done.
     */
    Content[row][i + col] = MakePrintable(ch_assign[i]);
  }

} /* SetRange */

/**
 * Set the content at the first column of a given row on this Page.
 * Copies the given String until either end of row or end of string.
 *
 * @param row 0-based row index of the String to modify.
 * @param to_copy String to copy at the given row.
 */
public void
SetRange(int row, String to_copy)
{
  char[]        ch_assign;

  ch_assign = to_copy.toCharArray();
  for (int i = 0; i < MAX_COLUMNS; i++) {
    /*
     * NEEDSWORK: same comments as above.
     */
    Content[row][i] = ch_assign [i];
  }

} /* SetRange */

/*------------------------------------------------------------------------
 *                      PUBLIC ACCESSORS
 */
/**
 * Returns the content of self at a given row-column location.
 *
 * @param row 0-based row index of character to return.
 * @param col 0-based column index of character to return.
 *
 * @return The character at the given row-column location.
 */ 
public char
GetChar(int row, int col)
{
  return (Content[row][col]);

} /* GetChar */

/**
 * Get the width of the widest row on this Page.  This number will
 * never be larger than the MAX_COLUMNS constant.
 *
 * @return The widest row on this Page.
 */
public int
GetMaxRowWidth()
{
  return (Max_row_width);

} /* GetMaxRowWidth */

/**
 * Get the number of rows present on this Page.  This number will
 * never be larger than the MAX_ROWS constant.  Trailing blank rows
 * are not included in the number of rows.
 *
 * @return The number of rows in this Page.
 */
public int
GetNumberOfRows()
{
  return (Number_of_rows);

} /* GetNumberOfRows */

/**
 * Get the content at a given row-column position for a given length
 * from this Page.  The returned String is guaranteed to fit on a
 * single line.
 *
 * @param row 0-based row index of the String to return.
 * @param col 0-based column index of the String to return.
 * @param len Length of the String to return.
 *
 * @return The substring from this Page at the given location.
 */
public String
GetRange(int row, int col, int len)
{
  if (row >= Number_of_rows || col >= Row_width[row])
    return ("");

  if (col + len > Row_width[row])
    len = Row_width[row] - col;
  return (new String(Content[row], col, len));

} /* GetRange */

/**
 * Get the content at a given row-column position from this Page.
 * Returns the rest of the row starting at the given position.
 *
 * @param row 0-based row index of the String to return.
 * @param col 0-based column index of the String to return.
 *
 * @return The substring from this Page at the given location.
 */
public String
GetRange(int row, int col)
{
  if (row >= Number_of_rows || col >= Row_width[row])
    return ("");

  return (new String(Content[row], col, Row_width[row] - col));

} /* GetRange */

/**
 * Get the contents of a given row position on this Page.
 * Returns the entire row.
 *
 * @param row 0-based row index of the String to return.
 *
 * @return A String representing the contents of the given row.
 */
public String
GetRange(int row)
{
  if (row >= Number_of_rows)
    return ("");

  return (new String(Content[row], 0, Row_width[row]));

} /* GetRange */

/**
 * Get the length of a given row.
 *
 * @param row 0-based index of the row whose length is to be returned.
 *
 * @return The length of the given row.
 */
public int
GetRowWidth(int row)
{
  if (Debug_out != null)
    Debug_out.println("Called Page::GetRowWidth");

  return (Row_width[row]);

} /* GetRowWidth */

/**
 * Read self from a given DataInputStream.
 *
 * @param in The DataInputStream which will fill the page contents.
 *
 * @return True if we hit the end of the input file, false otherwise.
 */
public boolean
Read(DataInputStream din)
{
  char		ch;
  int		row = 0,
		col = 0;
  boolean	hit_eof = false;

  if (Debug_out != null)
    Debug_out.println("Called Page::Read");

  if (Number_of_rows > 0)
    Clear();

  try {
processing_loop:
    for (;;) {
      ch = (char)din.readByte();
      /*
       * Till end of input page ('\f'), end of file, or
       * output buffer is full (last line is full).
       */
      switch (ch) {
        case END_OF_ROW_CHAR:
	  // Advance one row
	  ++row;
	  if (row >= MAX_ROWS)
	    break processing_loop;
	  col = 0;
          break;
        case CARRIAGE_RETURN:
	  break;
        case END_OF_PAGE_CHAR:
	  break processing_loop;
        default:
          Content[row][col] = MakePrintable(ch);
          Row_width[row] = col + 1;
          if (Row_width[row] > Max_row_width)
            Max_row_width = Row_width[row];
          Number_of_rows = row + 1;
          if (++col >= MAX_COLUMNS) {
            col = 0;
            ++row;
            if (row >= MAX_ROWS)
	      break processing_loop;
          }
          break;
      }
    }
  } catch (EOFException eof) {
    // Normal end-of-file termination.
    hit_eof = true;
  } catch (IOException ioe) {
    // System.out.println("Page::Read Exception:");
    // System.err.println(ioe);
  }
  TrimTrailingSpace();
  return (hit_eof);

} /* Read */

/**
 * Returns the contents of the Page as a String.
 *
 * @return String The Page contents returned in a String form.
 */
public String
toString()
{
  String        str = "";

  if (Number_of_rows > 0 && Max_row_width > 0) {
    for (int i = 0; i < Number_of_rows; i++) {
      str += GetRange(i);
      if (i + 1 >= Number_of_rows)
        str += END_OF_PAGE_STR;
      else
        str += END_OF_ROW_STR;
    }
  }
  return (str);

} /* toString */

/**
 * Returns the contents of the Page as a String with no end of page
 * character (unlike toString()).
 *
 * @return String The Page contents returned in a String form.
 */
public String
Page_Text()
{
  String        str = "";

  if (Number_of_rows > 0 && Max_row_width > 0) {
    for (int i = 0; i < Number_of_rows; i++) {
      str += GetRange(i) + END_OF_ROW_STR;
    }
  }
  return (str);

} /* Page_Text */

/**
 * Write self onto a given OutputStream.
 *
 * @param out The OutputStream which will receive the page contents.
 *
 * @return OutputStream The contents of the page.
 */
public void
Write(OutputStream out)
{
  byte []	output  = new byte[1];

  try {
    for (int row = 0; row < Number_of_rows; row++) {
      for (int col = 0; col < Row_width[row]; col++) {
	output[0] = (byte)Content[row][col];
        out.write(output);
      }
      output[0] = END_OF_ROW_BYTE;
      out.write(output);
    }
    output[0] = END_OF_PAGE_BYTE;
    out.write(output);
  } catch (IOException ioe) {
    // System.out.println("Page::Write Exception:");
    // System.err.println(ioe);
  }

} /* Write */

/*========================================================================
 *                      PRIVATE METHODS
 */
/**
 * Copy the contents from a given Page into self.
 *
 * @param to_copy The Page to copy into self.
 */
private void
Copy(Page to_copy)
{
  for (int i = 0; i < MAX_ROWS; i++) {
    for (int j = 0; j < MAX_COLUMNS; j++) {
      Content[i][j] = to_copy.Content[i][j];
    }
    Row_width[i] = to_copy.Row_width[i];
  }
  Number_of_rows = to_copy.Number_of_rows;
  Max_row_width = to_copy.Max_row_width;

} /* Copy */

private char
MakePrintable(char ch)
{
  if (ch < ' ' || ch > '~') {
    if (Debug_out != null) {
      int	i = (int)ch;

      Debug_out.println("The following char was converted to a \'.\' : "
			+ ch + " (" + i + ")");
    }
    ch = '.';
  }
  return (ch);

} /* MakePrintable */

private void SetDebugOut(PrintWriter ps)
{
  Debug_out = ps;

} /* SetDebugOut */

/**
 * Trim trailing white space from all rows of this page.  Trim
 * trailing blank rows from this page.  Recompute maximum row
 * width.
 */
private void
TrimTrailingSpace()
{
  int	col,
  	new_max_width = 0,
  	row;

  for (row = 0; row < Number_of_rows; row++) {
    for (col = Row_width[row] - 1; col >= 0; col--) {
      if (!Character.isSpaceChar(Content[row][col]))
	break;
    }
    Row_width[row] = col + 1;
    if (Row_width[row] > new_max_width)
      new_max_width = Row_width[row];
  }
  for (row = Number_of_rows - 1; row >= 0; row --)
    if (Row_width[row] != 0)
      break;
  Number_of_rows = row + 1;
  Max_row_width = new_max_width;

} /* TrimTrailingSpace */

/**
 * Test the Page class.
 */
public static void
main(String [] args)
{
  if (args.length > 2 || args.length == 1) {
    System.out.println("Usage: java Page [infile outfile]");
  } else if (args.length == 2) {
    try {
      DataInputStream		in_rp = new DataInputStream(
      					new BufferedInputStream(
					new FileInputStream(args[0])));
      BufferedOutputStream	out_rp = new BufferedOutputStream(
					new FileOutputStream(args[1]));
      Page			page = new Page();
      boolean			hit_eof = false;

      System.out.print("Reading report file...");
      do {
	System.out.print(".");
        hit_eof = page.Read(in_rp);
	if (!hit_eof || page.GetNumberOfRows() > 0)
	  page.Write(out_rp);
	out_rp.flush();
      } while (!hit_eof);
      System.out.println("done.");

      in_rp.close();
      out_rp.close();

    } catch(FileNotFoundException fnf) {
      System.out.println(fnf);
      return;
    } catch(IOException ioe) {
      System.out.println(ioe);
      return;
    }
  } else {
    try {
      FileOutputStream	output	= new FileOutputStream ("errlog.txt");
      PrintWriter	out	= new PrintWriter(new
					  BufferedOutputStream(output));
      FileInputStream	in	= new FileInputStream ("report.txt");
      BufferedInputStream bin	= new BufferedInputStream(in);
      DataInputStream	stream	= new DataInputStream (bin);

      out.println("Testing default Page constructor.");

      Page		p1 = new Page();

      p1.SetDebugOut(out);
      out.println("Empty Page p1:");
      out.println("  Number_of_rows = " + p1.GetNumberOfRows());
      out.println("  Max_row_width = " + p1.GetMaxRowWidth());
      for (int i = 0; i < p1.GetNumberOfRows(); i++)
	out.println("  Row_width[" + i + "] = " + p1.GetRowWidth(i));
      out.println("<" + p1.toString() + ">");
      out.println("");
      out.println("Testing DataInputStream-based Page constructor.");

      Page		p2 = new Page(stream);

      p2.SetDebugOut(out);
      out.println("Stream Page p2:");
      out.println("  Number_of_rows = " + p2.GetNumberOfRows());
      out.println("  Max_row_width = " + p2.GetMaxRowWidth());
      for (int i = 0; i < p2.GetNumberOfRows(); i++)
	out.println("  Row_width[" + i + "] = " + p2.GetRowWidth(i));
      out.println("<" + p2.toString() + ">");
      out.println("");
      out.println("Testing Copy-based Page constructor.");
      Page		p3 = new Page(p2);
      p3.SetDebugOut(out);
      out.println("Copy Page p3:");
      out.println("  Number_of_rows = " + p3.GetNumberOfRows());
      out.println("  Max_row_width = " + p3.GetMaxRowWidth());
      for (int i = 0; i < p3.GetNumberOfRows(); i++)
	out.println("  Row_width[" + i + "] = " + p3.GetRowWidth(i));
      out.println("<" + p3.toString() + ">");

      out.close();
      stream.close();
    } catch(FileNotFoundException fnf) {
      System.out.println(fnf);
      return;
    } catch(IOException ioe) {
      System.out.println(ioe);
      return;
    }
  }

} /* main */

} /* class Page */

/*========================================================================
 *                      END OF FILE $RCSfile: Page.java,v $
 */
